/***************************************************************************/
/* 		This code is part of Nscache - viewer of Netscape(tm)	   */
/*		browsers disk cache					   */
/*		Copyright (c) 1999,2000 Ondrejicka Stefan		   */
/*		(ondrej@idata.sk)					   */
/*		created and modified 2007 ... 2012 by Harald Foerster	   */
/*		(harald_foerster@users.sourceforge.net)			   */
/*		Distributed under GPL 2 or later			   */
/***************************************************************************/

#include <errno.h>
#include <limits.h>
#include <string.h>

#include "charconv.h"
#include "gui-gtk.h"

/*
#include <glib.h>
#include "nls.h"
*/

#if GLIB_MAJOR_VERSION == 1

#include <iconv.h>
#include <stdlib.h>

#if defined(BUILD_CYGWIN) && defined(HAVE_LOCALE)
#ifndef HAVE_GETTEXT
#include <locale.h>
#endif
#endif

#ifdef HAVE_LANGINFO
#include <langinfo.h>
#endif

#ifdef HAVE_UNICODE
#include <unicode.h>
#endif

#define CHARCONV_T iconv_t

#define CHARCONV_OPEN(to, from) \
		iconv_open((to), (from))
#define CHARCONV_RUN(cd, instr, inlen, outstr, outlen) \
		iconv((cd), (instr), (inlen), (outstr), (outlen))
#define CHARCONV_CLOSE(cd) \
		iconv_close(cd)

#else /* GLIB_MAJOR_VERSION > 1 */

#include "nscache.h"

#define CHARCONV_T GIConv

#define CHARCONV_OPEN(to, from) \
		g_iconv_open((to), (from))
#define CHARCONV_RUN(cd, instr, inlen, outstr, outlen) \
		g_iconv((cd), (instr), (inlen), (outstr), (outlen))
#define CHARCONV_CLOSE(cd) \
		g_iconv_close(cd)

#endif /* #if GLIB_MAJOR_VERSION == 1 .. #else */


#ifndef ICONV_CONST
#define ICONV_CONST
#endif


const gchar charconv_iso8859_1_str[]	= "ISO-8859-1";
const gchar charconv_utf8_str[]		= "UTF-8";
const gchar *charconv_locale_str;


gchar *charconv(const gchar *str, const gchar *to, const gchar *from, gint *tflen)
{
	size_t			sz_in;
	CHARCONV_T		cd;
	ICONV_CONST gchar	*buf_in;

	buf_in = (gchar *) str;

	if (str == NULL)
	{
		return NULL;
	}

	if (tflen == NULL || (sz_in = *tflen) < 0)
	{
		sz_in = strlen(str);
	}

	if (sz_in == 0)
	{
		goto __nothing_to_do__;
	}

	if (to == NULL)
	{
		to = charconv_iso8859_1_str;
	}

	if (from == NULL)
	{
		from = charconv_iso8859_1_str;
	}

	if (to == from || GuiStrCaseCmp(to, from) == 0)
	{
__nothing_to_do__:
		if (tflen && *tflen < 0)
		{
			*tflen = sz_in;
		}

		return (gchar *) buf_in;
	}

	cd = CHARCONV_OPEN(to, from);

	if (cd != (CHARCONV_T) -1)
	{
		gchar	*buf_out, *retv;
		gint	result;
		size_t	sz_out, length;

		length  = sz_in;
		sz_out  = 0;
		retv    = NULL;
		buf_out = NULL;

		do
		{
			size_t next;

			next     = buf_out - retv;
			length <<= 2;
			retv     = g_realloc(retv, length + 1);
			buf_out  = &retv[next];
			sz_out   = length - next;

			result = CHARCONV_RUN(cd, &buf_in, &sz_in, &buf_out, &sz_out);
		}
		while (result < 0 && errno == E2BIG);

		CHARCONV_CLOSE(cd);

		if (sz_in == 0)
		{
			length -= sz_out;
			retv[length] = '\0';

			if (tflen)
			{
				*tflen = length;
			}

			return g_realloc(retv, length + 1);
		}

		g_free(retv);
	}

	return NULL;
}

#if GLIB_MAJOR_VERSION == 1

void charconv_init(void)
{
	/*
		'nl_langinfo' under Cygwin
		or 'unicode_get_charset'
		seems always returning an
		'ASCII-US' character set
	*/

#ifdef HAVE_LANGINFO

	gchar *charset;
#endif

#ifdef HAVE_UNICODE

	unicode_init();
#endif

#ifdef HAVE_LANGINFO

#ifdef BUILD_CYGWIN
#ifdef HAVE_LOCALE

	charset = setlocale(LC_CTYPE, "de_DE");

	if (charset)
	{
		setlocale(LC_CTYPE, "");
		charset = nl_langinfo(CODESET);
	}
#else
	charset = NULL;
#endif

#else
	charset = nl_langinfo(CODESET);
#endif
	if (charset)
	{
		if (GuiStrCaseCmp(charset, charconv_iso8859_1_str) == 0)
		{
			charconv_locale_str = charconv_iso8859_1_str;
		}

		else
		{
			charconv_locale_str = g_strdup(charset);
		}
	}
	else

#endif /* HAVE_LANGINFO */

	charconv_locale_str = "";
}

#else /* GLIB_MAJOR_VERSION > 1 */

void charconv_init(void)
{
	g_get_charset(&charconv_locale_str);

	if (strcmp(charconv_locale_str, charconv_iso8859_1_str) == 0)
	{
		charconv_locale_str = charconv_iso8859_1_str;
	}

	else
	{
		charconv_locale_str = g_strdup(charconv_locale_str);
	}
}

gchar *charconv_utf8_from_locale(const gchar *str)
{
	if (str && g_utf8_validate(str, -1, NULL) == FALSE)
	{
		return g_locale_to_utf8(str, -1, NULL, NULL, NULL);
	}

	return (gchar *) str;
}

gchar *charconv_utf8_from_locale_dup(const gchar *str)
{
	gchar *utf = charconv_utf8_from_locale(str);

	if (utf == str)
	{
		return g_strdup(utf);
	}

	return utf;
}

gint charconv_utf8_from_locale_snprintf(gchar *sbuf, gint slen, const gchar *fmt, ...)
{
	gint	len;
	va_list	args;

	va_start(args, fmt);
	len = charconv_utf8_from_locale_vsnprintf(sbuf, slen, fmt, args);
	va_end(args);

	return len;
}

gint charconv_utf8_from_locale_vsnprintf(gchar *sbuf, gint slen, const gchar *fmt, va_list args)
{
#if ULONG_MAX < UINT_MAX
	static gchar format_str[] = "%lu";
	gulong	num;
#else
	static gchar format_str[] = "%u";
	guint	num;
#endif

	gint chr, len, sz;

	len = sz = 0;

	while ((chr = *(fmt++)) != '\0')
	{
		gchar	*str, *utf;

		if (chr != '%')
		{
			sbuf[len] = chr;
			sz = 1;
		}

		else switch (*(fmt++))
		{
			case 'd':
			case 'u':
				num = va_arg(args, guint);
				goto __do_snprintf__;

			case 'l':
				num = va_arg(args, gulong);
__do_snprintf__:
				sz  = g_snprintf(&sbuf[len], slen, format_str, num);

				if (*fmt == 'u')
				{
					fmt++;
				}

				break;

			case 's':
				str = va_arg(args, gchar*);

				if (str == NULL || *str == '\0')
				{
					continue;
				}

				utf = CHARCONV_UTF8_FROM_LOCALE(str);

				if (utf == NULL)
				{
					gchar *s = g_strdup(str);
					utf = tl_replace_non_ascii_chars_in_str(s);
				}

				sz  = g_snprintf(&sbuf[len], slen, "%s", utf);
				CHARCONV_UTF8_FREE(utf, str);
				break;

			default:
				sz = -1;
				break;
		}

		if (sz < 0 || (len + sz) >= slen)
		{
			sz = -1;
			break;
		}

		len  += sz;
		slen -= sz;
	}

	sbuf[len] = '\0';

	return (sz < 0) ? sz : len;
}

gchar *charconv_utf8_to_locale(const gchar *str)
{
	if (str && g_utf8_validate(str, -1, NULL))
	{
		return g_locale_from_utf8(str, -1, NULL, NULL, NULL);
	}

	return (gchar *) str;
}

gchar *charconv_utf8_to_locale_dup(const gchar *str)
{
	gchar *utf = charconv_utf8_to_locale(str);

	if (utf == str)
	{
		return g_strdup(utf);
	}

	return utf;
}

gchar *charconv_utf8_gettext(const gchar *str)
{
	if (str)
	{
		gchar *utf, *nls;

		nls = gettext(str);
		utf = charconv_utf8_from_locale(nls);

		if (utf != str)
		{
			if (utf == nls)
			{
				return g_strdup(utf);
			}

			if (utf != NULL)
			{
				return utf;
			}
		}
	}

	return (gchar *) str;
}

gchar *charconv_utf8_from_filename(const gchar *str, gint *len)
{
	if (str)
	{
		if (tl_is_ascii_str(str) == FALSE)
		{
#ifdef BUILD_MINGW
			return g_locale_to_utf8(str, -1, NULL, len, NULL);
#else
			return g_filename_to_utf8(str, -1, NULL, len, NULL);
#endif
		}

		if (len)
		{
			*len = strlen(str);
		}
	}

	return (gchar *) str;
}

gchar *charconv_utf8_to_filename(const gchar *str)
{
	if (str && g_utf8_validate(str, -1, NULL))
	{
#ifdef BUILD_MINGW
		return g_locale_from_utf8(str, -1, NULL, NULL, NULL);
#else
		return g_filename_from_utf8(str, -1, NULL, NULL, NULL);
#endif
	}

	return (gchar *) str;
}

#endif /* #if GLIB_MAJOR_VERSION == 1 .. #else */

/* EOF */
