/***************************************************************************/
/* 		This code is part of Nscache - viewer of Netscape(tm)	   */
/*		browsers disk cache					   */
/*		Copyright (c) 1999,2000 Ondrejicka Stefan		   */
/*		(ondrej@idata.sk)					   */
/*		created and modified 2005 ... 2012 by Harald Foerster	   */
/*		(harald_foerster@users.sourceforge.net)			   */
/*		Distributed under GPL 2 or later			   */
/***************************************************************************/

#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>

#include "gui-gtk.h"
#include "file.h"
#include "moz.h"

#ifdef HAVE_ZLIB
#include "zlib.h"
#ifdef HAVE_BZLIB
#include "bzlib.h"
#endif
#else
#ifdef HAVE_SYS_WAIT_H
#include <signal.h>
#include <sys/wait.h>
#endif
#endif

#ifdef HAVE_CYGWIN_CONV_PATH
#include <sys/cygwin.h>
#endif

#ifdef HAVE_MSIE
#include "msie.h"
#endif

/*
#include "nscache.h"
#include "stringbuf.h"
#include <sys/types.h>
#include <sys/stat.h>
*/


typedef struct
{
	const nscache_record_t*	nsc_rec;
	size_t			url_len;
	const char*		url_str;
	size_t			ufn_len;
	const char*		ufn_str;
	size_t			dfn_len;
	char*			dfn_str;
	size_t			dst_len;
	char			dst_str[1];

} URLinfo;


/* globals needed in 'gui.c' function 'SaveFolderAs()' */

const char index_html_str[]	= "/index.html";

unsigned int decode;
unsigned int single_folder = 0;
unsigned int url_to_local_selector = 0;

#ifdef HAVE_LINK
unsigned int hardlink = 1;
#endif


#if defined(BUILD_CYGWIN) || defined(BUILD_MINGW)
#define WINDOWS_FILENAME 1
#endif


#ifdef WINDOWS_FILENAME

static void file_name_replace(char* fname)
{
	/*
		MS Windows doesn't like the following
		characters (without '/' and '\')
	*/

	static const char replace_char[] = "?:*|<>\"";

	const char* ptr = replace_char;

	do
	{
		char* s = fname - 1;

		while((s = strchr(++s, *ptr)) != NULL)
		{
			*s = '!';
		}
	}
	while(*(++ptr) != '\0');

} /* static void file_name_replace(char*) */

static void file_path_replace(char* fpath)
{
	/* Skip eg. "C:/" else converted to "C!/"*/

	char *s = strchr(fpath, '/');

	file_name_replace(s == NULL ? fpath : s);
}

#endif /* WINDOWS_FILENAME */

#ifdef BUILD_MINGW

void file_path_dos_to_unix(char *path)
{
	while ((path = strchr(path, '\\')) != NULL)
	{
		*(path++) = '/';
	}
}

void file_path_unix_to_dos(char *path)
{
	while ((path = strchr(path, '/')) != NULL)
	{
		*(path++) = '\\';
	}
}

#endif /* BUILD_MINGW */

static int file_make_pathname(const char* filename)
{
	int	slen;
	char*	str;

	slen = nscache_db.path.len;

	str = stringbuf_append(&nscache_db.path, filename, -1);

	if(str == NULL)
	{
		return -1;
	}

	return slen;

} /* static int file_make_pathname(const char*) */

#ifdef HAVE_LINK

static int file_link(const nscache_record_t* rec, char* dst)
{
	int slen, retv;;

	if(rec->content_offset)
	{
		return -1;
	}

	slen = file_make_pathname(rec->filename);

	if(slen == -1)
	{
		return slen;
	}

#ifdef WINDOWS_FILENAME
	file_path_replace(dst);
#endif
	retv = link(nscache_db.path.str, dst);

	stringbuf_truncate(&nscache_db.path, slen);

	return retv;

} /* static int file_link(const nscache_record_t*, char*) */

#endif /* HAVE_LINK */

int file_open(const char* filename, int flags)
{
	int fd, slen;

	slen = file_make_pathname(filename);

	if(slen == -1)
	{
		return slen;
	}

	fd = open(nscache_db.path.str, flags);

	stringbuf_truncate(&nscache_db.path, slen);

	return fd;

} /* int file_open(const char*, int) */

#ifndef CACHE_READONLY

int file_remove(const char* filename)
{
	int slen, retv;

	slen = file_make_pathname(filename);

	if(slen == -1)
	{
		return slen;
	}

	retv = unlink(nscache_db.path.str);

	stringbuf_truncate(&nscache_db.path, slen);

	return retv;

} /* int file_remove(const char*) */

int file_rmdir(const char* dirname)
{
	int slen, retv;

	slen = file_make_pathname(dirname);

	if(slen == -1)
	{
		return slen;
	}

	retv = rmdir(nscache_db.path.str);

	stringbuf_truncate(&nscache_db.path, slen);

	return retv;

} /* int file_rmdir(const char*) */

#endif /* CACHE_READONLY */

static int file_access_rdwr(const char* path)
{
	int error = errno;

	if(access(path, R_OK | W_OK))
	{
		int retv = FILE_ACCESS_READ;

		if(access(path, R_OK))
		{
			error = errno;
			retv  = access(path, F_OK);
		}

		if(retv >= 0)
		{
			errno = error;
		}

		return retv;
	}

	return FILE_ACCESS_READ | FILE_ACCESS_WRITE;

} /* static int file_access_rdwr(const char*) */

int file_access_index(void)
{
	int slen, retv;

	slen = file_make_pathname(nscache_db.name);

	if(slen == -1)
	{
		return slen;
	}

	retv = file_access_rdwr(nscache_db.path.str);

	if(retv == 0)
	{
		int error = errno;

		if(access(nscache_db.path.str, W_OK) == 0)
		{
			/* write only ? */
			retv = FILE_ACCESS_WRITE;
		}

		errno = error;
	}

	stringbuf_truncate(&nscache_db.path, slen);

	return retv;

} /* int file_access_index(void) */

int file_access(const nscache_record_t* rec)
{
	int slen, retv;

	if(rec->content_offset)
	{

#ifdef HAVE_MSIE
		if(rec->content_offset == MSIE_CACHE)
		{
			return msie_file_access(rec);
		}
#endif

		retv = moz_file_access(rec);

#ifndef CACHE_READONLY
		if(retv < 0)
#endif
			return retv;
	}

	/* cache file */

	slen = file_make_pathname(rec->filename);

	if(slen == -1)
	{
		return slen;
	}

	retv = file_access_rdwr(nscache_db.path.str);

#ifndef CACHE_READONLY
	if(retv > 0 && (retv & FILE_ACCESS_WRITE))
	{
		/* cache directory */

		char *sep;

		if(rec->content_offset)
		{
			goto index_access;
		}

		sep = strrchr(nscache_db.path.str, '/');

		if(sep)
		{
			int acc;

			*sep = '\0';
			 acc = file_access_rdwr(nscache_db.path.str);
			*sep = '/';

			if(acc > 0 && (acc & FILE_ACCESS_WRITE))
			{
index_access:
				stringbuf_truncate(&nscache_db.path, slen);

				/* index file */

				slen = file_make_pathname(nscache_db.name);

				if(slen == -1)
				{
					return retv;
				}

				acc = file_access_rdwr(nscache_db.path.str);

				if(acc == (FILE_ACCESS_READ | FILE_ACCESS_WRITE))
				{
					retv |= FILE_ACCESS_REMOVE;
				}
			}
		}
	}
#endif /* CACHE_READONLY */

	stringbuf_truncate(&nscache_db.path, slen);

	return retv;

} /* int file_access(const nscache_record_t*) */

int file_chmod(const nscache_record_t* rec, mode_t mode)
{
	int slen, retv;

	if(rec->content_offset)
	{
		return -1;
	}

	slen = file_make_pathname(rec->filename);

	if(slen == -1)
	{
		return slen;
	}

	retv = chmod(nscache_db.path.str, mode);

	stringbuf_truncate(&nscache_db.path, slen);

	return retv;

} /* int file_chmod(const nscache_record_t*, mode_t) */

int file_status(const nscache_record_t* rec, struct stat* statbuf)
{
	int slen, retv;

	if(rec->content_offset)
	{
		return -1;
	}

	slen = file_make_pathname(rec->filename);

	if(slen == -1)
	{
		return slen;
	}

	retv = stat(nscache_db.path.str, statbuf);

	stringbuf_truncate(&nscache_db.path, slen);

	return retv;

} /* int file_status(const nscache_record_t*, struct stat*) */

#ifdef HAVE_ZLIB

static void* file_decode(const char *encoding, const char *buffer, size_t *length)
{

#ifdef HAVE_BZLIB
	char**		next_out;
	unsigned int*	avail_out;
	unsigned int*	total_out;
#else

#define next_out	(&(u.zs.next_out))
#define avail_out	(&(u.zs.avail_out))
#define total_out	(&(u.zs.total_out))

#endif

	int		status, winbits;
	char*		pbuf;

	union
	{
		z_stream	zs;
#ifdef HAVE_BZLIB
		bz_stream	bzs;
#endif
	} u;

	pbuf = NULL;

	if(strcmp(encoding, "gzip") == 0 ||
		strcmp(encoding, "deflate") == 0)
	{
		winbits = 15 + 32;

#ifdef HAVE_BZLIB
		next_out  = (char**) &(u.zs.next_out);
		avail_out = (unsigned int*) &(u.zs.avail_out);
		total_out = (unsigned int*) &(u.zs.total_out);
	}

	else if(strcmp(encoding, "bzip2") == 0)
	{
		winbits   = 0;
		next_out  = (char**) &(u.bzs.next_out);
		avail_out = (unsigned int*) &(u.bzs.avail_out);
		total_out = (unsigned int*) &(u.bzs.total_out_lo32);
#endif
	}

	else
	{
		return pbuf;
	}

	for(;;)
	{
		memset(&u, '\0', sizeof(u));

#ifdef HAVE_BZLIB
		if(winbits == 0)
		{
			u.bzs.next_in	= (char *) buffer;
			u.bzs.avail_in	= *length;
			status		= BZ2_bzDecompressInit(&u.bzs, 0, 0);
		}

		else

#endif /* HAVE_BZLIB */

		{
			u.zs.next_in	= (char *) buffer;
			u.zs.avail_in	= *length;
			status		= inflateInit2(&u.zs, winbits);
		}

		if(status == 0)
		{
			for(;;)
			{
				char* ptr;

				*avail_out = *length << 2;

				ptr  = pbuf;
				pbuf = realloc(pbuf, *avail_out + *total_out);

				if(pbuf == NULL)
				{
					if(ptr)
					{
						free(ptr);
					}

					break;
				}

				*next_out = &pbuf[*total_out];
#ifdef HAVE_BZLIB
				if(winbits == 0)
				{
					status = BZ2_bzDecompress(&u.bzs);

					if(status == BZ_STREAM_END)
					{
						goto _stream_end;
					}
				}
				else
#endif
				if((status = inflate(&u.zs, Z_SYNC_FLUSH)) == Z_STREAM_END)
				{
#ifdef HAVE_BZLIB
_stream_end:
#endif
					ptr  = pbuf;
					pbuf = realloc(pbuf, *total_out + 1);

					if(pbuf == NULL)
					{
						free(ptr);
						break;
					}

					pbuf[*total_out] = '\0';
					*length = *total_out;
					break;
				}

				if(status != 0)
				{
					free(pbuf);
					pbuf = NULL;
					break;
				}

			} /* for(;;) */
#ifdef HAVE_BZLIB
			if(winbits == 0)
			{
				BZ2_bzDecompressEnd(&u.bzs);
				break;
			}
#endif
			inflateEnd(&u.zs);

			if(pbuf != NULL)
			{
				break;
			}

		} /* if(status == 0) */

		if(winbits < 0)
		{
			break;
		}

		winbits = -15;

	} /* for(;;) */

	return pbuf;
}

#else /* ! HAVE_ZLIB */

#ifdef HAVE_SYS_WAIT_H

static void* file_decode(const char* encoding, const char* buffer, size_t* length)
{
	int		fdpin[2], fdpout[2];
	char*		pbuf;
	pid_t		child_exec;
	sigset_t	sig_block, sig_unblock;

	pbuf = NULL;

	if(strcmp(encoding, "gzip") &&
			strcmp(encoding, "bzip2"))
	{
		return pbuf;
	}

	if(pipe(fdpin))
	{
		return pbuf;
	}

	if(pipe(fdpout))
	{
		close(fdpin[0]);
		close(fdpin[1]);

		return pbuf;
	}

	/* block SIGCHLD (in fact all signals) */
	sigfillset(&sig_block);
	sigprocmask(SIG_BLOCK, &sig_block, &sig_unblock);

	switch(child_exec = fork())
	{
		case  0:
		{
			close(fdpin[1]);
			close(fdpout[0]);

			if(dup2(fdpin[0], STDIN_FILENO) >= 0 &&
					dup2(fdpout[1], STDOUT_FILENO) >= 0)
			{
				/* "gzip -cdq < buffer" */
				execlp(encoding, encoding, "-cdq", NULL);
			}

			_exit(-1);

		} /* case  0 (child_exec) */

		case -1:
		{
			close(fdpin[0]);
			close(fdpin[1]);
			close(fdpout[0]);
			close(fdpout[1]);
			break;
		}

		default:
		{
			pid_t child_write;

			close(fdpin[0]);
			close(fdpout[1]);

			switch(child_write = fork())
			{
				case  0:
				{
					int rlen  = 0;
					int count = *length;

					do
					{
						int retv = write(fdpin[1], &buffer[rlen], count);

						if(retv < 0)
						{
							_exit(retv);
						}

						rlen  += retv;
						count -= retv;
					}
					while(count > 0);

					_exit(0);

				} /* case  0 (child_write) */

				case -1:
				{
					close(fdpin[1]);
					close(fdpout[0]);
					break;
				}

				default:
				{
					int count, plen, rlen, retv, status;

					close(fdpin[1]);

					count = plen = rlen = 0;

					for(;;)
					{
						if(count == 0)
						{
							char* ptr;

							count = *length << 2;
							plen += count;
							ptr   = pbuf;
							pbuf  = realloc(pbuf, plen);

							if(pbuf == NULL)
							{
								free(ptr);
								break;
							}
						}

						retv = read(fdpout[0], &pbuf[rlen], count);

						if(retv == 0)
						{
							break;
						}

						if(retv < 0)
						{
							free(pbuf);
							pbuf = NULL;
							break;
						}

						rlen  += retv;
						count -= retv;

					} /* for(;;) */

					close(fdpout[0]);
					waitpid(child_write, &status, 0);

					if(pbuf != NULL)
					{
						if(WIFEXITED(status) == FALSE ||
								WEXITSTATUS(status) != 0)
						{
							free(pbuf);
							pbuf = NULL;
						}

						else
						{
							char* ptr = pbuf;

							pbuf = realloc(pbuf, rlen + 1);

							if(pbuf == NULL)
							{
								free(ptr);
							}

							else
							{
								pbuf[rlen] = '\0';
								*length = rlen;
							}
						}

					} /* if(pbuf != NULL) */

					break;

				} /* default (parent_write) */

			} /* switch(child_write = fork()) */

			waitpid(child_exec, NULL, 0);

			break;

		} /* default (parent_exec) */

	} /* switch(child_exec = fork()) */

	/* unblock signals */
	sigprocmask(SIG_SETMASK, &sig_unblock, NULL);

	return pbuf;

} /* static void* file_decode(const char*, const char*, size_t*) */

#else

#define file_decode(encoding, buffer, length)	NULL

#endif /* HAVE_SYS_WAIT_H */

#endif /* #ifdef HAVE_ZLIB .. #else */

static void* file_real_read(int fd, const nscache_record_t* rec, size_t* length)
{
	int	offset;
	char*	buffer;

	if(rec->content_offset)
	{
		*length = rec->content_length;

		offset = lseek(fd, rec->content_offset, SEEK_SET);

		if(offset != rec->content_offset)
		{
			if(offset != -1)
			{
				errno = ENODATA;
			}

			return NULL;
		}
	}

	else
	{
		*length = offset = lseek(fd, 0, SEEK_END);

		if(offset == -1 || (offset = lseek(fd, 0, SEEK_SET)) != 0)
		{
			if(offset != -1)
			{
				errno = ENODATA;
			}

			return NULL;
		}
	}

	buffer = malloc(*length + 1);

	if(buffer)
	{
		int len;

		buffer[*length] = '\0';

		len = read(fd, buffer, *length);

		if(len != *length)
		{
			if(len != -1)
			{
				errno = ENODATA;
			}

			free(buffer);
			return NULL;
		}

		if(rec->content_encoding && decode)
		{
			char *ptr;

			ptr = file_decode(rec->content_encoding, buffer, length);

			if(ptr)
			{
				free(buffer);
				return ptr;
			}
		}
	}

	return buffer;

} /* static void* file_real_read(int, const nscache_record_t*, size_t*) */

void* file_read(const nscache_record_t* rec, size_t* length)
{
	int	fd;
	char*	buffer;

	if(rec->content_offset)
	{
#ifdef HAVE_MSIE
		if(rec->content_offset == MSIE_CACHE)
		{
			return msie_file_read(rec, length);
		}
#endif
		if(moz_file_access(rec) < 0)
		{
			return NULL;
		}
	}

	fd = file_open(rec->filename, FILE_OPEN_READ_ONLY);

	if(fd == -1)
	{
		return NULL;
	}

	buffer = file_real_read(fd, rec, length);

	close(fd);

	return buffer;

} /* void* file_read(const nscache_record_t*, size_t*) */

void* file_blk_read(const nscache_record_t* rec, size_t* length)
{
	int fd;

	if(rec->content_offset == 0)
	{
		return file_read(rec, length);
	}

	fd = moz_blk_access(rec);

	if(fd < 0)
	{
		return NULL;
	}

	return file_real_read(fd, rec, length);

} /* void* file_blk_read(const nscache_record_t*, size_t*) */

static int file_write(char* dst, mode_t file_perm, const char* buffer, size_t length)
{
	int fd, len;

#ifdef WINDOWS_FILENAME
	file_path_replace(dst);
#endif
	fd = open(dst, FILE_OPEN_WRITE_ONLY, file_perm);

	if(fd == -1)
	{
#ifdef BUILD_MINGW
		/* errno was set to 'ENOENT' instead 'ENAMETOOLONG' */
		if(errno == ENOENT && strlen(dst) > PATH_MAX - 1)
		{
			errno = ENAMETOOLONG;
		}
#endif
		return FILE_ERR_OPEN_DST;
        }

	len = write(fd, buffer, length);

	close(fd);

	if(len != length)
	{
		unlink(dst);

		if(len != -1)
		{
			errno = ENOSPC;
		}

		return FILE_ERR_WRITE;
	}

	return 0;

} /* static int file_write(char*, mode_t, const char*, size_t) */

int file_copy(const nscache_record_t* rec, char* dst, mode_t file_perm)
{
	int	error;
	char*	buffer;
	size_t	length;

	buffer = file_read(rec, &length);

	if(buffer == NULL)
	{
		return FILE_ERR_OPEN_SRC;
	}

	error = file_write(dst, file_perm, buffer, length);

	free(buffer);

	return error;

} /* int file_copy(const nscache_record_t*, char*, mode_t) */

int file_save(const nscache_record_t* rec, char* dst)
{

#ifdef HAVE_LINK

	/*
		Try to link this file only if it's 'stand alone',
		and it should'nt be decoded, and the user have
		decided so.
		If it's a Mozilla data cache file, we have to read
		the bytes from the first (which is the offset) to
		the last  (which is the offset+length) byte.
		Files from MS Internet Explorer will always be readed.
	*/

	if(hardlink && rec->content_offset == 0 &&
			(rec->content_encoding == NULL || decode == FALSE))
	{
		if(file_link(rec, dst) == 0)
		{
			return 0;
		}

		if(errno != EPERM && errno != EXDEV)
		{
			return FILE_ERR_LINK;
		}
	}

#endif /*  HAVE_LINK */

	return file_copy(rec, dst, FILE_PERM_USER);

} /* int file_save(nscache_record_t*, char*) */

static int directory_create(const char* path)
{
	int    count, error;
	char*  last_sep;

	/*
		We do create the folders recursively, because doing
		so is very easy. Each time this function is called,
		we first shrinken the path by replacing the last se-
		parator with the string terminator. The first time
		we hide the file name, after then all folders that
		doesn't exist yet. This function is called by itself
		until the first (most top) folder can be created, then
		each instance of this function will add its part (sub
		folder) until getting back to the first instance, where
		the whole path will be completed by creating the last
		(most bottom) sub folder.
	*/

	last_sep = strrchr(path, '/');

	if(last_sep == NULL)
	{
		return FILE_ERR_DIR_CREATE;
	}

	while(last_sep > path && last_sep[-1] == '/')
	{
		last_sep--;
	}

	*last_sep = '\0';
	count	  = 0;

#ifdef HAVE_UNIX_MKDIR
	while((error = mkdir(path, DIR_PERMISSION)) != 0)
#else
	while((error = mkdir(path)) != 0)
#endif
	{
		error = errno;

		/*
			The only error we can accept is that the dest-
			ination doesn't exist (that's why we're here).
		*/

		if(error == ENOENT)
		{
			error = 0;

			count = directory_create(path);

			if(count == FILE_ERR_DIR_CREATE)
			{
				error = count;
			}
		}

		if(error)
		{
			break;
		}
	}

	/* don't forget to restore the name */

	*last_sep = '/';

	if(error == 0)
	{
		/* summery of created folders */
		return ++count;
	}

	return FILE_ERR_DIR_CREATE;

} /* static int directory_create(const char*) */

static int is_index_html(const char* url_fname)
{
	/*
		A name with no dot is treated as folder, with
		one or more dot(s) or capital letter(s) as file.
		This can cause errors, e.g.
			readme		file treated as folder
			README		folder treated as file
			folder.img	folder treated as file
	*/

	unsigned int chr;

	while( (chr = *(url_fname++)) != '\0' )
	{
		/* is a file name */

		if(chr == '.' || chr == '#' ||
			chr == '?' ||  isupper(chr))
		{
			return FALSE;
		}
	}

	/* append 'index.html' */

	return TRUE;

} /* static int is_index_html(const char*) */

static int file_convert(const URLinfo* iptr, const URLinfo** info_array,
						size_t service_len, StringBuf* filebuf)
{
	size_t converted, file_offs, buf_start;

	buf_start = file_offs = converted = 0;

	do
	{
		unsigned int ch;

		/* find start of instruction */

		ch = filebuf->str[file_offs];

		if(ch == '\0')
		{
			continue;
		}

		if(ch == '<')
		{
			file_offs++;
			ch = filebuf->str[file_offs];

			if(ch == '/' || ch == '>')
			{
				buf_start = 0;
				continue;
			}

			if(ch == '!' && filebuf->str[file_offs + 1] == '-' &&
						filebuf->str[file_offs + 2] == '-')
			{
				char* rend;

				/* skip "!--" */
				file_offs += 3;

				rend = tl_mem_find_str(&filebuf->str[file_offs],
							filebuf->len - file_offs, "-->", 3);
				if(rend == NULL)
				{
					return converted;
				}

				/* skip "--" */
				file_offs = &rend[2] - filebuf->str;
				buf_start = 0;

				continue;
			}

			buf_start = file_offs;
		}

		else if(ch == '>' && buf_start > 0)
		{
			static const struct
			{
				char*	str;
				size_t	len;

			} *link_ptr, link_array[] =
				{
					{"href=",	5},
					{"src=",	4},
					{"longdesc=",	9},
					{"action=",	7},
					{"background=",11},
					{NULL,		0}
				};

			static const char updir_str[] = "../";
 
			char*		url_start;
			char*		url_keep;
			const char*	url_name;
			char*		s;
			int		diff, updir_count;
			size_t		url_start_offs;
			size_t		url_keep_offs;
			size_t		tmp_end_offs;
			size_t		ulen;
			const URLinfo* 	cache;
			const URLinfo** downloaded;

			struct
			{
				size_t	offs;
				int	chr;

			} url_end;

			struct
			{
				size_t	offs;
				size_t	len;
				int	chr;

			} url_fnam;


			if(file_offs - buf_start < 5)
			{
				/* min. "src= >" */
				continue;
			}

			if(file_offs - buf_start > 10 + service_len)
			{
				if(GuiStrNCaseCmp(&filebuf->str[buf_start],
								"base href=", 10) == 0)
				{
					/* start of remark ("<!-- <") */

					if(stringbuf_insert(filebuf, "!-- <", 5,
									buf_start) == NULL)
					{
						/* out of memory */
						return -1;
					}

					/* end of remark ("> -->") */

					if(stringbuf_insert(filebuf, "> --", 4,
									file_offs + 5) == NULL)
					{
						/* out of memory */
						return -1;
					}

					converted++;
					file_offs += 9;
					buf_start  = 0;

					continue;

				} /* if(GuiStrNCaseCmp(&filebuf->str[buf_start] ...) == 0) */

			} /* if(file_offs - buf_start > 11 + service_len) */


			/* now start conversion ... */

			link_ptr = link_array;

			while((url_start = (char*) link_ptr->str) != NULL)
			{
				url_start =
				tl_mem_find_str_igncase(&filebuf->str[buf_start],
					file_offs - buf_start, link_ptr->str, link_ptr->len);

				if(url_start)
				{
					break;
				}

				link_ptr++;

			} /* while((url_start = (char*) link_ptr->str) != NULL) */


			if(url_start == NULL ||
				( url_start[-1] != '<' && ! isspace(url_start[-1]) ))
			{
				goto __loop_continue__;
			}

			filebuf->str[file_offs] = '\0';
			url_start += link_ptr->len;

			if(*url_start == '\"')
			{
				ch = *(url_start++);
				s  = strchr(url_start, ch);

				/* not really neccessary - but safety first */

				if(s == NULL)
				{
					goto __loop_continue__;
				}
			}

			else
			{
				s = url_start;

				do
				{
					ch = *(++s);
				}
				while(ch != '\0' && ! isspace(ch));
			}

			url_start_offs	= url_start - filebuf->str;
			url_end.offs	= s - filebuf->str;
			url_end.chr 	= ch;

			/* terminate url */
			*s = '\0';

			if(strncmp(url_start, iptr->url_str, service_len) == 0)
			{
				/* service found - skip it */

				url_keep_offs = url_start_offs + service_len;

				if(filebuf->str[url_keep_offs] == '/')
				{
					url_keep_offs++;
				}

				else if(filebuf->str[url_keep_offs] != '\0')
				{
					/* differs - e.g. "foo" vs. "foo_1" */
					goto __urlend_restore__;
				}
			}

			else if(*url_start == '/')
			{
				/* relative addressing mode */
				url_keep_offs = url_start_offs + 1;
			}

			else if(*url_start == '#')
			{
				/* nothing to do */
				goto __urlend_restore__;
			}

			else if(strstr(url_start, "://"))
			{
				/* other service */
				goto __urlend_restore__;
			}

			else
			{
				/* relative to current folder */
				url_keep_offs = url_start_offs;

			} /* if(strncmp(url_start, iptr->url_str, service_len) == 0) */


			/*
				Compare the URL with the links for matching
				the complete URL, or whether it is a folder
				and needs appending 'index.html'.
			*/

			s = &filebuf->str[url_keep_offs - 1];

			/* skip leading slash */

			while(*(++s) == '/')
			{
				url_keep_offs++;
			}

			url_keep = s;
			s = &filebuf->str[url_end.offs];
			url_name = tl_get_url_filename(url_keep);
			ulen = s - url_keep;
			url_fnam.chr = '\0';

			if(url_name == NULL)
			{
				/* last character is a slash */
				url_name = s;
			}

			url_fnam.offs = url_name - filebuf->str;
			url_fnam.len  = s - url_name;

			while(s > url_name && (ch = *(--s)) != '/')
			{
				if(ch == '#')
				{
					url_fnam.len = s - url_name;
					url_fnam.chr = ch;
					ulen = s - url_keep;
					break;
				}
			}

			downloaded = info_array;

			while((cache = *(downloaded++)) != NULL)
			{
				size_t		clen, tlen;
				const char	*sc, *st;

				st   = url_keep;
				sc   = &cache->url_str[service_len];
				tlen = ulen;

				while(*(++sc) == '/') {;}

				clen = &cache->url_str[cache->url_len] - sc;

				if(url_keep_offs == url_start_offs)
				{
					/*
						URL is relative to current
						folder, skip "./" and "../"
						Compare path with the cache
						and the current URL.
					*/

					const char *beg = &iptr->url_str[service_len];
					const char *end =  iptr->ufn_str;

					/* skip leading slash */
					while(*(++beg) == '/') {;}

					if(st[0] == '.')
					{
						/* last slash */
						end--;

						while(st[0] == '.' && st < url_name)
						{
							if(st[1] == '/')
							{
								st++;
							}

							else if(st[1] == '.' && st[2] == '/')
							{
								if(beg >= end)
								{
									break;
								}

								st += 2;

								/* find slash */
								while(*(--end) != '/') {;}

								/* skip slash */
								while(*(--end) == '/') {;}

								/* points to slash */
								end++;
							}

							else
							{
								break;
							}

							/* skip slash */
							while(*(++st) == '/') {;}

						} /* while(st[0] == '.' && st < url_name) */

						/* skip slash */
						while(*(++end) == '/') {;}

						if(end < beg)
						{
							/* path don't match */
							continue;
						}

						tlen -= st - url_keep;

					} /* if(st[0] == '.') */

					/* compare includes last slash */

					if(strncmp(sc, beg, end - beg))
					{
						/* path don't match */
						continue;
					}

					clen -= end - beg;
					sc    = &cache->url_str[cache->url_len - clen];

				} /* if(url_keep_offs == url_start_offs) */

				if(tlen == clen)
				{
					/*
						url_keep		cache
						--------		-----
						"dir/"			"dir/"
						"dir/name"		"dir/name"
						"name"			"name"
						"name.html#somewhat"	"name.html"
					*/

					if(tlen == 0 || strncmp(sc, st, tlen) == 0)
					{
						/* file was downloaded */

						break;
					}
				}

				else if(tlen == clen - 1 && url_fnam.chr != '#')
				{
					/*
						url_keep		cache
						--------		-----
						"dir"			"dir/"
					*/

					if(cache->ufn_str[0] == '\0' &&
						(tlen == 0 || strncmp(sc, st, tlen) == 0))
					{
						/* "index.html" was downloaded */

						url_fnam.offs = url_end.offs;
						url_fnam.len  = 0;

						break;
					}
				}

				else
				{
					/*
						Test if either the URL or the
						cache contains the 'index.html'.

						url_keep		cache
						--------		-----
						"dir"			"dir/index.html"
						"dir/"			"dir/index.html"
						"dir/index.html"	"dir/"
						"dir/index.html#some"	"dir/"
					*/

					size_t min, plen;

					if(clen < tlen)
					{
						/* cache without "index.html" */
						min  = clen;
						plen = tlen - clen;
					}

					else
					{
						/* URL without "index.html" */
						min  = tlen;
						plen = clen - tlen;

						if(url_name[0] == '\0')
						{
							/* exclude last slash */
							plen--;
						}
					}

					if(plen != sizeof(index_html_str) - 2)
					{
						/* not "index.html" */
						continue;
					}

					if(min == 0 || strncmp(sc, st, min) == 0)
					{
						if(min == clen)
						{
							/* cache without "index.html" */
							sc = url_name;
							st = cache->ufn_str;
						}

						else
						{
							/* URL without "index.html" */
							sc = cache->ufn_str;
							st = url_name;
						}

						if(st[0] == '\0' &&
							strncmp(sc, &index_html_str[1],
								sizeof(index_html_str) - 2) == 0)
						{
							/* "index.html" was downloaded */

							if(cache->ufn_str[0] != '\0' &&
									url_name[0] != '\0' &&
										url_fnam.chr != '#')
							{
								/* URL without "index.html" */

								url_fnam.offs = url_end.offs;
								url_fnam.len  = 0;
							}

							break;
						}
					}
				}

			} /* while((cache = *(downloaded++)) != NULL) */

			if(cache == NULL)
			{
				if(url_to_local_selector == URL_CONVERT_ONLY_FILE_REFERENCE)
				{
					/* file was not downloaded */
					goto __urlend_restore__;
				}

				if(url_fnam.len > 0 &&
					is_index_html(url_name) &&
						tl_mem_find_str_igncase(url_keep,
							url_name - url_keep, "cgi-bin/", 8) == NULL)
				{
					url_fnam.offs = url_end.offs;
					url_fnam.len  = 0;
				}
			}

			updir_count = 0;

			if(single_folder)
			{
				/* skip last slash */
				url_keep_offs = url_fnam.offs;
			}

			else if(url_keep_offs > url_start_offs)
			{
				const char*	url_dest;
				char*		url_link;

				/* skip "http://www.foo.bar/" (including slash) */

				url_dest = &(iptr->url_str[service_len + 1]);
				url_link = url_keep;

				/* remaining folders plus filename */

				while(url_dest < iptr->ufn_str &&
					(s = strchr(url_dest, '/')) != NULL)
				{
					/* compare includes last slash */

					if(strncmp(url_dest, url_link, &s[1] - url_dest))
					{
						break;
					}

					url_link += s - url_dest;

					while(*(++url_link) == '/') {;}

					url_dest = s;

					while(*(++url_dest) == '/') {;}
				}

				url_keep_offs += url_link - url_keep;

				while(url_dest < iptr->ufn_str &&
					(s = strchr(url_dest, '/')) != NULL)
				{
					while(*(++s) == '/') {;}

					/* climb one folder up */
					updir_count++;

					url_dest = s;
				}

			} /* if(single_folder) .. else if(url_keep_offs > url_start_offs) */

			/* move down => neg. value */

			diff	 = url_start_offs - url_keep_offs;
			diff	+= updir_count * (sizeof(updir_str) - 1);

			if((diff | updir_count) != 0)
			{
				char* ptr;

				if(stringbuf_move(filebuf, url_start_offs, diff) == NULL)
				{
					/* out of memory */
					return -1;
				}

				ptr = &filebuf->str[url_start_offs];

				while(--updir_count >= 0)
				{
					/* prepend "../" */
					ptr  = memcpy(ptr, updir_str, sizeof(updir_str) - 1);
					ptr += sizeof(updir_str) - 1;
				}

			} /* if((diff | updir_count) != 0) */

			tmp_end_offs   = url_end.offs + diff;
			url_fnam.offs += diff;

			url_keep = &filebuf->str[url_fnam.offs];
			url_name = NULL;

			if(cache)
			{
				/* Replace URL filename with the disk filename */

				if(url_fnam.len != cache->dfn_len ||
						strncmp(url_keep, cache->dfn_str, cache->dfn_len))
				{

					size_t url_nlen, slen;

					url_name = cache->dfn_str;
					url_nlen = cache->dfn_len;

					if(url_fnam.len == 0)
					{
						if(tmp_end_offs > url_start_offs &&
									url_keep[-1] != '/')
						{
							/* '/index.html' */
							url_nlen++;
							url_name--;
						}
					}

					else if(url_fnam.chr == '#')
					{
						tmp_end_offs = url_fnam.offs + url_fnam.len;
					}

					slen = url_nlen - url_fnam.len;

					if(slen)
					{
						diff += slen;

						if(stringbuf_move(filebuf,
								tmp_end_offs, slen) == NULL)
						{
							/* out of memory */
							return -1;
						}
					}

					/* replace filename */
					memcpy(url_keep, url_name, url_nlen);
				}
			}

			else if(url_fnam.len == 0)
			{
				/* append 'index.html' */

				size_t url_nlen;

				url_name = &index_html_str[1];
				url_nlen = sizeof(index_html_str) - 2;

				if(tmp_end_offs > url_start_offs && url_keep[-1] != '/')
				{
					/* filename treated as folder, append slash */
					url_nlen++;
					url_name--;
				}

				diff += url_nlen;

				if(stringbuf_insert(filebuf,
						url_name, url_nlen, tmp_end_offs) == NULL)
				{
					/* out of memory */
					return -1;
				}

			} /* if(cache) .. else if(url_fnam.len == 0) */

			if(diff || updir_count || url_name)
			{
#ifdef WINDOWS_FILENAME
				file_name_replace(&filebuf->str[url_start_offs]);
#endif
				url_end.offs	+= diff;
				file_offs	+= diff;
				converted++;

			} /* if(diff || updir_count || url_name) */

__urlend_restore__:
			filebuf->str[url_end.offs] = url_end.chr;

__loop_continue__:
			filebuf->str[file_offs] = '>';

			buf_start = 0;

		} /* else if(ch == '>' && buf_start > 0) */
	}
	while(++file_offs < filebuf->len);

	return converted;

} /* static int file_convert(const URLinfo*, const URLinfo**, size_t, StringBuf*) */

static void url_info_free(URLinfo** info_array)
{
	URLinfo** ihdl;
	URLinfo*  iptr;

	ihdl = info_array;

	while((iptr = *(ihdl++)) != NULL)
	{
		free(iptr);
	}

	free(info_array);
}

static URLinfo** url_info_new(const char* dst,
				const nscache_record_t** rec_array, int depth,
					SaveDirResult* sdr, size_t* service_length)
{
	int*		fnum;
	size_t		count, dir_length;
	const char*	service_beg;
	const char*	service_end;
	URLinfo**	info_array;
	URLinfo**	ihdl;
	StringBuf	path;

	const nscache_record_t** recs;
	const nscache_record_t*  cache;

#ifdef BUILD_CYGWIN
	int	winlen;
#ifdef HAVE_CYGWIN_CONV_PATH
	char	buf[PATH_MAX];
	cygwin_conv_to_full_win32_path("/cygdrive", buf);
	winlen = strlen(buf);
#else
	winlen = sizeof("C:/cygwin/cygdrive") - 1;
#endif
#endif

	if(rec_array == NULL ||
		*rec_array == NULL ||
			stringbuf_init(&path, dst, -1) == NULL)
	{
		/* out of memory */
		errno = ENOMEM;

		return NULL;
	}

	if(access(path.str, F_OK) == 0)
	{
		/* folder exists */
		stringbuf_delete(&path);
		errno = EEXIST;

		return NULL;
	}

	dir_length = path.len;

	cache = *rec_array;

	/* length of e.g. "http://www.foo.bar" */

	if((service_beg = tl_get_url_start(cache->urlstr)) == NULL ||
		(service_end = strchr(&(strchr(service_beg, ':'))[3], '/')) == NULL)
	{
		/* invalid argument */
		stringbuf_delete(&path);
		errno = EINVAL;

		return NULL;
	}

	/* length without trailing slash */

	*service_length = service_end - service_beg;

	recs = rec_array;

	while(*(recs++) != NULL) {;}

	count = recs - rec_array;

	info_array = calloc(count, sizeof(URLinfo*));

	if(info_array == NULL)
	{
		stringbuf_delete(&path);
		return NULL;
	}

	fnum = calloc(count - 1, sizeof(int));

	if(fnum == NULL)
	{
		free(info_array);
		stringbuf_delete(&path);
		return NULL;
	}

	ihdl = info_array;
	recs = rec_array;

	while((cache = *(recs++)) != NULL)
	{
		unsigned int	chr;
		char*		dst;
		char*		urlstr;
		char*		urlfname;
		size_t		flen, start;
		URLinfo*	iptr;
		URLinfo**	itmp;

		stringbuf_truncate(&path, dir_length);

		urlstr = tl_get_url_start(cache->urlstr);

		if(urlstr == NULL)
		{
			sdr->errors++;
			continue;
		}

		dst = &urlstr[*service_length];

		/*
			If we've got something like this
			"http://www.foo.bar/load.html?http://www.foo.bar/index.html"
			then the URL will changed to
			"http://www.foo.bar/load.html?http:--www.foo.bar-index.html"
		*/

		urlfname = tl_get_url_filename(dst);
		chr = '\0';

		if(urlfname)
		{
			chr = *urlfname;
			*urlfname = '\0';
		}

		if(single_folder)
		{
			/*
				If writing the files into a single
				directory (without any sub directories)
				we need only the file name.
			*/

			dst = strrchr(dst, '/');
		}

		else
		{
			int   loop;
			char* s;

			if(*dst != '/')
			{
				sdr->errors++;
				continue;
			}

			loop = depth;

			while(--loop >= 0)
			{
				if(dst == NULL)
				{
					sdr->errors++;
					continue;
				}

				/* skip all folders that won't be saved */

				while(*(++dst) == '/') {;}

				dst = strchr(dst, '/');
			}

			for(;;)
			{
				while(*(++dst) == '/') {;}

				s = strchr(dst, '/');

				/* previous slash */
				dst--;

				if(s == NULL)
				{
					break;
				}

				if(stringbuf_append(&path, dst, s - dst) == NULL)
				{
					dst = NULL;
					break;
				}

				dst = s;
			}

		} /* if(single_folder) .. else */

		if(dst == NULL || stringbuf_append(&path, dst, -1) == NULL)
		{
			sdr->errors++;
			continue;
		}

		start = path.len;

		if(urlfname)
		{
			char* s;

			*urlfname = chr;

			if(stringbuf_append(&path, urlfname, -1) == NULL)
			{
				sdr->errors++;
				continue;
			}

			s = &path.str[start];

			/* maybe URL in URL */

			while((s = strchr(++s, '/')) != NULL)
			{
				*s = '-';
			}
		
			if(strchr(&path.str[start], '.') == NULL)
			{
				stringbuf_append(&path,
					tl_get_url_extension(cache->content_type), -1);
			}
		}

		else
		{
			stringbuf_append(&path, &index_html_str[1],
							sizeof(index_html_str) - 2);
		}

#ifdef WINDOWS_FILENAME

		/*
			Truncate path AND filename to 'PATH_MAX',
			which includes the terminating '\0'
		*/

		flen = path.len;
#ifdef BUILD_CYGWIN
		flen += winlen;
#endif

		if(flen > PATH_MAX - 1)
		{
			flen = PATH_MAX - 1;
#ifdef BUILD_CYGWIN
			flen -= winlen;
#endif
			stringbuf_truncate(&path, flen);
		}
#else
		/*
			Truncate filename to 'NAME_MAX',
			if neccessary (sometimes too long)
		*/

		flen = path.len - start;

		if(flen > NAME_MAX)
		{
			flen = NAME_MAX;
			stringbuf_truncate(&path, start + flen);
		}
#endif

		/* Look for already existing filenames */

		itmp = info_array;

		while((iptr = *(itmp++)) != NULL)
		{
			if(iptr->dst_len == path.len &&
						strcmp(iptr->dst_str, path.str) == 0)
			{
				int  len;
				int* pnum;
				char buf[16];

				pnum = &fnum[itmp - info_array - 1];

				(*pnum)++;
				len = sprintf(buf, "_X%d", *pnum);

				if(len > 0)
				{
					int	dif;
					size_t	pos;
					char*	dot;

					dot = strrchr(&path.str[start], '.');

					if(dot)
					{
						pos = dot - path.str;
					}

					else
					{
						pos = path.len;
					}

#ifdef WINDOWS_FILENAME
					dif = flen + len - (PATH_MAX - 1);
#ifdef BUILD_CYGWIN
					dif += winlen;
#endif
#else
					dif = flen + len - NAME_MAX;
#endif
					if(dif > 0)
					{
						pos -= dif;
						stringbuf_truncate(&path,
								path.len - dif);
					}

					stringbuf_insert(&path, buf, len, pos);
				}

				break;
			}

		} /* while((iptr = *(itmp++)) != NULL) */

		iptr = malloc(sizeof(URLinfo) + path.len);

		if(iptr == NULL)
		{
			url_info_free(info_array);
			info_array = NULL;
			break;
		}

		*(ihdl++) = iptr;

		iptr->nsc_rec = cache;
		iptr->url_len = strlen(urlstr);
		iptr->url_str = urlstr;

		if(urlfname == NULL)
		{
			iptr->ufn_len = 0;
			iptr->ufn_str = &(iptr->url_str[iptr->url_len]);
		}

		else
		{
			iptr->ufn_len = strlen(urlfname);
			iptr->ufn_str = urlfname;
		}

		iptr->dfn_len = path.len - start;
		iptr->dfn_str = &(iptr->dst_str[start]);
		iptr->dst_len = path.len;
		strcpy(iptr->dst_str, path.str);

	} /* while((cache = *(recs++)) != NULL) */

	if(fnum)
	{
		free(fnum);
	}

	stringbuf_delete(&path);

	return info_array;

} /* static URLinfo* url_info_new(const char*, const nscache_record_t**, int,
							SaveDirResult*, size_t*) */

int folder_save(const char* dst,
		const nscache_record_t** rec_array, int depth, SaveDirResult* sdr)
{
	int		error;
	size_t		service_length;
	URLinfo**	info_array;
	URLinfo**	ihdl;
	URLinfo*	iptr;

	memset(sdr, '\0', sizeof(SaveDirResult));

	info_array = url_info_new(dst, rec_array, depth, sdr, &service_length);

	if(info_array == NULL)
	{
		sdr->errors++;
		return FILE_ERR_DIR_CREATE;
	}

	ihdl = info_array;

	while(error = 0, (iptr = *(ihdl++)) != NULL)
	{
		int		converted;
		StringBuf	filebuf;

		converted = 0;

		if(url_to_local_selector != URL_CONVERT_NONE &&
			(iptr->nsc_rec)->content_encoding == NULL &&
					(iptr->nsc_rec)->content_type &&
					strcmp((iptr->nsc_rec)->content_type, "text/html") == 0)
		{
			char*	buffer;
			size_t	length;

			buffer = file_read(iptr->nsc_rec, &length);

			if(buffer)
			{
				stringbuf_from_string(&filebuf, buffer, length);

				converted = file_convert(iptr, (const URLinfo**) info_array,
								service_length, &filebuf);
				if(converted <= 0)
				{
					stringbuf_delete(&filebuf);

					if(converted < 0)
					{
						/* ENOMEM */
						sdr->errors++;
					}
				}

			} /* if(buffer) */

		} /* if(url_to_local_selector != URL_CONVERT_NONE) */

		while( (error = ( (converted > 0) ?
			file_write(iptr->dst_str, FILE_PERM_USER, filebuf.str, filebuf.len) :
					file_save(iptr->nsc_rec, iptr->dst_str) )) != 0 )
		{

#ifdef HAVE_LINK
			if(error == FILE_ERR_LINK)
			{
				if(file_access(iptr->nsc_rec) == -1)
				{
					error = FILE_ERR_OPEN_SRC;
					break;
				}
			}
			else
#endif
			if(error != FILE_ERR_OPEN_DST)
			{
				break;
			}

			/*
				This type of error is for us essential:
					ENOENT: folder doesn't exists.

				If we've got another error,
				we'll continue with the next file.
			*/

			if(errno != ENOENT)
			{
				break;
			}

			error = directory_create(iptr->dst_str);

			if(error == FILE_ERR_DIR_CREATE)
			{
				break;
			}

			sdr->folders += error;

		} /* while( (error = ( (converted > 0) ? ... */

		if(converted > 0)
		{
			stringbuf_delete(&filebuf);
		}

		if(error)
		{
			sdr->errors++;

			if(error == FILE_ERR_DIR_CREATE)
			{
				break;
			}

			if(sdr->errnr == 0)
			{
				sdr->errnr = errno;

				if(error == FILE_ERR_OPEN_SRC)
				{
					sdr->erec = iptr->nsc_rec;
				}

				else
				{
					sdr->fname = malloc(iptr->dst_len + 1);

					if(sdr->fname)
					{
						sdr->fname = strcpy(sdr->fname, iptr->dst_str);
					}
				}

			} /* if(sdr->errnr) */
		}

		else
		{
			sdr->files++;
			sdr->converted += converted;

		} /* if(error) .. else */

	} /* while(error = 0, (iptr = *(ihdl++)) != NULL) */

	url_info_free(info_array);

	return error;

} /* int folder_save(const char* dst,
		const nscache_record_t**, int, SaveDirResult*) */

/* EOF */
