/***************************************************************************/
/* 		This code is part of Nscache - viewer of Netscape(tm)	   */
/*		browsers disk cache					   */
/*		Copyright (c) 1999,2000 Ondrejicka Stefan		   */
/*		(ondrej@idata.sk)					   */
/*		modified 2005 ... 2009,2012 by Harald Foerster		   */
/*		(harald_foerster@users.sourceforge.net)			   */
/*		Distributed under GPL 2 or later			   */
/***************************************************************************/

#include "nscache.h"
#include "file.h"

#include <ctype.h>
#include <errno.h>
#include <string.h>

#ifdef CONFIG_FNAME_COMPLETION

#ifdef HAVE_GLOB
#include <glob.h>
#elif defined(HAVE_FNMATCH)
#include <fnmatch.h>
#else
#include "fnmatch.h"
#endif

#ifdef BUILD_MINGW
#include "charconv.h"
#endif

#endif /* CONFIG_FNAME_COMPLETION */

/*
#include <glib.h>
#include "stringbuf.h"
*/

#ifndef PRE_INCREMENT
#ifdef __powerpc__
#define PRE_INCREMENT		1
#endif
#endif

#define kAsciiLineFeed		0x0a
#define kAsciiCarriageReturn	0x0d

int tl_is_ascii_str(const char *str)
{
	unsigned int		c;
	const unsigned char	*p;

#ifdef PRE_INCREMENT
	p = str - 1;
	while ((c = *(++p)) != '\0')
#else
	p = str;
	while ((c = *(p++)) != '\0')
#endif
	{
		if (c > 0x7f)
		{
			return FALSE;
		}
	}

	return TRUE;

} /* int tl_is_ascii_str(const char*) */

int tl_is_alnum_str(const char *beg, const char *str, int slen)
{
	if ( (( str == beg || isalnum(str[-1]) == FALSE ) &&
					isalnum(str[slen]) == FALSE) )
	{
		return TRUE;
	}

	return FALSE;

} /* int tl_is_alnum_str(const char*, const char*, int) */

int tl_count_lines(const char *p, const char *end)
{
	int count = 1;

#ifdef PRE_INCREMENT
	p--;
	end--;
#endif

	while (end > p)
	{
		unsigned int c;

#ifdef PRE_INCREMENT
		c = *(++p);
#else
		c = *(p++);
#endif
		if (c == kAsciiLineFeed)
		{
			/* Unix */
			count++;
		}

		else if (c == kAsciiCarriageReturn)
		{
			/* Mac ... */

			if (end > p)
			{
#ifdef PRE_INCREMENT
				c = p[1];
#else
				c = p[0];
#endif
				if (c == kAsciiLineFeed)
				{
					/* ... or DOS */
					p++;
				}
			}

			count++;
		}
	}

	return count;

} /* int tl_count_lines(const char*, const char*) */

char *tl_strndup_to_lower(const char *str, int slen)
{
	char *ptr;

	if (slen < 0)
	{
		slen = strlen(str);
	}

	ptr = g_malloc(slen + 1);
	ptr[slen] = '\0';

	while (--slen >= 0)
	{
		unsigned int c = str[slen];

		ptr[slen] = tolower(c);
	}

	return ptr;

} /* char *tl_strndup_to_lower(const char*, int) */

char *tl_replace_non_ascii_chars_in_str(char *str)
{
	unsigned int	c;
	unsigned char	*p;

#ifdef PRE_INCREMENT
	p = str - 1;
	while ((c = *(++p)) != '\0')
	{
		if (c > 0x7f)
		{
			p[0] = '?';
		}
	}
#else
	p = str;
	while ((c = *(p++)) != '\0')
	{
		if (c > 0x7f)
		{
			p[-1] = '?';
		}
	}
#endif
	return str;

} /* char *tl_replace_non_ascii_chars_in_str(char*) */

char *tl_get_linestart(const char *beg, const char *p)
{
	while (p > beg)
	{
		unsigned int c = *(--p);

		if (c < 0x80 && isprint(c) == FALSE)
		{
			return (char *) p + 1;
		}
	}

	return (char *) beg;

} /* char *tl_get_linestart(const char*, const char*) */

char *tl_get_linestop(const char *p, const char *end)
{
	unsigned int c;

#ifdef PRE_INCREMENT
	p--;
	end--;
	while (end > p)
	{
		c = *(++p);

		if (c < 0x80 && isprint(c) == FALSE)
		{
			return (char *) p;
		}
	}

	return (char *) end + 1;
#else
	while (end > p)
	{
		c = *(p++);

		if (c < 0x80 && isprint(c) == FALSE)
		{
			return (char *) p - 1;
		}
	}

	return (char *) end;
#endif

} /* char *tl_get_linestop(const char*, const char*) */

static char *tl_mem_find_init(const char *mem, int mlen, const char *str, int *slen)
{
	if (mem == NULL || str == NULL ||
				mlen == 0 || *slen == 0)
	{
		return NULL;
	}

	if (mlen < 0)
	{
		mlen = strlen(mem);
	}

	if (*slen < 0)
	{
		*slen = strlen(str);
	}

	mlen -= *slen;

	if (mlen < 0)
	{
		return NULL;
	}

	return (gchar *) &mem[mlen];

} /* static char *tl_mem_find_init(const char*, int, const char*, int*) */

char *tl_mem_find_str(const char *mem, int mlen, const char *str, int slen)
{
	unsigned int	firstc, c;
	const char	*stop;

	stop = tl_mem_find_init(mem, mlen, str, &slen);

	if (stop == NULL)
	{
		return NULL;
	}

	firstc = *str;

#ifdef PRE_INCREMENT

	mem--;

	while (c = *(++mem), stop >= mem)
#else
	stop++;

	while (c = *(mem++), stop >= mem)
#endif
	{
		if (c == firstc)
		{
			const unsigned char *s1, *s2;
#ifdef PRE_INCREMENT
			s1 = &mem[slen];
#else
			s1 = &mem[slen - 1];
#endif
			s2 = &str[slen];

			do
			{
				c = *(--s2);

				if (str == (const char *) s2)
				{
#ifdef PRE_INCREMENT
					return (char *) mem;
#else
					return (char *) mem - 1;
#endif
				}

			}
			while (c == *(--s1));

		} /* if (c == firstc) */

	} /* while (c = *(mem++), stop >= mem) */

	return NULL;

} /* char *tl_mem_find_str(const char*, int, const char*, int) */

char *tl_mem_find_str_igncase(const char *mem, int mlen, const char *str, int slen)
{
	unsigned int	firstc, c;
	const char	*stop;

	stop = tl_mem_find_init(mem, mlen, str, &slen);

	if (stop == NULL)
	{
		return NULL;
	}

	firstc = *str;

#ifdef PRE_INCREMENT

	mem--;

	while (c = tolower(*(++mem)), stop >= mem)
#else
	stop++;

	while (c = tolower(*(mem++)), stop >= mem)
#endif
	{
		if (c == firstc)
		{
			const unsigned char *s1, *s2;
#ifdef PRE_INCREMENT
			s1 = &mem[slen];
#else
			s1 = &mem[slen - 1];
#endif
			s2 = &str[slen];

			do
			{
				c = *(--s2);

				if (str == (const char *) s2)
				{
#ifdef PRE_INCREMENT
					return (char *) mem;
#else
					return (char *) mem - 1;
#endif
				}

			}
			while (c == tolower(*(--s1)));

		} /* if (c == firstc) */

	} /* while (c = *(mem++), stop >= mem) */

	return NULL;

} /* char *tl_mem_find_str_igncase(const char*, int, const char*, int) */

char *tl_get_mime_value_str(char *mem, int mem_len,
					const char *mime_id, int mid_len)
{
	unsigned char *beg, *stop;

	if (mem == NULL)
	{
		return NULL;
	}

	stop = &mem[mem_len];
	mem++;

	while ((beg =
		tl_mem_find_str_igncase(mem, (char *) stop - (char *) mem,
							mime_id, mid_len)) != NULL)
	{
		if (iscntrl(beg[-1]) || isspace(beg[-1]))
		{
			beg += mid_len;

			while ( stop > beg && (isprint(*beg) == FALSE || isspace(*beg)) )
			{
				beg++;
			}

			if (isalpha(*beg))
			{
				unsigned char *end = beg;

				while (stop > end && isprint(*end))
				{
					if (end[1] == ';')
					{
						break;
					}

					end++;
				}

				while (isspace(*end))
				{
					end--;
				}

				end[1] = '\0';

				return beg;

			} /* if (isalpha(*beg)) */

		} /* if (iscntrl(beg[-1]) || isspace(beg[-1])) */

		mem = &beg[1];

	} /* while ((beg = tl_mem_find_str_igncase(..)) != NULL) */

	return NULL;

} /* char *tl_get_mime_value_str(char*, int, const char*, int) */

char *tl_get_url_start(const char *urlstr)
{
	/*
		If we've got something like this
		"id=123456b&uri=http://www.foo.bar/index.php"
		then we'll skip the first part of the URL and use
		"http://www.foo.bar/index.php"
	*/

	char *beg = NULL;

	if (urlstr && (beg = strstr(urlstr, "://")))
	{
		while (beg > urlstr && (isalnum(beg[-1]) || beg[-1] == '-'))
		{
			beg--;
		}
	}

	return beg;

} /* char *tl_get_url_start(const char*) */

char *tl_get_url_filename(const char *urlstr)
{
	/*
		If we've got something like this
		"http://www.foo.bar/load.html?http://www.foo.bar/index.html"
		then the file name will set to
		"load.html?http://www.foo.bar/index.html"
	*/

	unsigned int	chr;
	char		*beg, *end, *str;

	if (urlstr == NULL || urlstr[0] == '\0')
	{
		return NULL;
	}

	end = (char *) &urlstr[strcspn(urlstr, "?;")];

	 chr = *end;
	*end = '\0';

	if (*urlstr == '/')
	{
		/* relative addressing mode */
		beg = (char *) urlstr;
	}

	else if((beg = strstr(urlstr, "://")) != NULL)
	{
		/* service like 'http://' - skip ':/' */
		beg += 2;
	}

	else
	{
		/* relative to current folder */
		beg = (char *) urlstr - 1;
	}

	while (*(++beg) == '/') {;}

	str = NULL;

	if (*beg != '\0' || chr != '\0')
	{
		str = strrchr(beg, '/');

		if (str == NULL)
		{
			str = beg;
		}

		else if (*(++str) == '\0')
		{
			str = NULL;
		}
	}

	*end = chr;

	return str;

} /* char *tl_get_url_filename(const char*) */

char *tl_get_url_extension(const char *mime)
{
	if (mime != NULL)
	{
		struct UrlExtension
		{
			const char *mime;
			const char *extension;
		};

		static const struct UrlExtension url_extension_tab[] =
		{
			{ "application/javascript",	".js"  },
			{ "application/js",		".js"  },
			{ "application/json",		".json"},
			{ "application/x-javascript",	".js"  },
			{ "application/xml",		".xml" },
			{ "image/gif",			".gif" },
			{ "image/jpeg",			".jpg" },
			{ "image/jpg",			".jpg" },
			{ "image/png",			".png" },
			{ "image/x-icon",		".ico" },
			{ "image/x-png",		".png" },
			{ "text/javascript",		".js"  },
			{ "text/x-javascript",		".js"  },
			{ "text/x-js",			".js"  },
			{NULL}
		};

		const struct UrlExtension *ptr = url_extension_tab;

		do
		{
			if (g_strcasecmp(ptr->mime, mime) == 0)
			{
				return (char *) ptr->extension;
			}

			ptr++;
		}
		while(ptr->mime != NULL);
	}

	return ".html";

} /* char *tl_get_url_extension(const char*) */

int tl_str_escape_chars(char *buf, int len, const char *str, const char *esc_chars)
{
	int	chr;
	char	*beg, *end;

	if (len <= 0)
	{
		return -1;
	}

	if (esc_chars == NULL)
	{
		esc_chars = "\0";
	}

	chr = '\0';
	beg = buf;

	if (str)
	{
		end = &buf[len - 1];

		while ((chr = *(str++)) != '\0' && buf < end)
		{
			if (strchr(esc_chars, chr))
			{
				*(buf++) = '\\';

				if (buf >= end)
				{
					break;
				}
			}

			*(buf++) = chr;
		}
	}

	*buf = '\0';

	return (chr == '\0') ? buf - beg : -1;

} /* int tl_str_escape_chars(char*, int, const char*, const char*) */

char *tl_get_1qstr(char *str)
{
	static char *ptr = NULL;

	if (str == NULL)
	{
		str = ptr;
	}

	if (str)
	{
		unsigned int chr;

		char *beg = NULL;

		while ( (chr = *(str++)) != '\0' )
		{
			if (chr =='\\')
			{
				memmove(&str[-1], str, strlen(str) + 1);
			}

			else if (chr =='\"')
			{
				if (beg)
				{
					str[-1] = '\0';
					ptr = str;
					return beg;
				}

				beg = str;
			}
		}
	}

	ptr = NULL;

	return NULL;

} /* char *tl_get_1qstr(char*) */

#ifdef CONFIG_FNAME_COMPLETION

#ifndef HAVE_GLOB

static void tl_sort_strarray(char **buf, int n)
{
	if (buf != NULL && n > 1)
	{
		char **upper = &buf[n];

		while (--upper > buf) /* &buf[1] */
		{
			char **lower = upper;

			while (--lower >= buf) /* &buf[0] */
			{
				if (g_strcasecmp(*lower, *upper) > 0)
				{
					char *tmp;

					tmp	= *upper;
					*upper	= *lower;
					*lower	= tmp;
				}
			}
		}

	} /* if (buf != NULL && n > 1) */

} /* static void tl_sort_strarray(char**, int) */

#define TL_FNMATCH_NAMES 2048

static int tl_fnmatch_completion(char *path, CompletionData *data)
{
	static const char *dir_separator = G_DIR_SEPARATOR_S;

	int		i;
	char		*pattern, *buf[TL_FNMATCH_NAMES + 1];
	const char	*fname;
	DIRdata		*dir;

#ifdef HAVE_OPENDIR
	struct dirent	*d;
#endif

	data->fnamec = 0;

	pattern = strrchr(path, G_DIR_SEPARATOR);

	if (pattern == NULL)
	{
		pattern = path;
		path = ".";
	}

	else if (pattern == path)
	{
		/* root directory */
		pattern++;
		path = (char *) dir_separator;
	}

#ifdef BUILD_MINGW

	else if (pattern == &path[2] && path[1] == ':')
	{
		/* drive eg "C:\*" */
		pattern = memmove(&path[4], &path[3], strlen(pattern) + 1);
		path[3] = '\0';
	}
#endif
	else
	{
		/* terminate pathname */
		*(pattern++) = '\0';
	}

	dir = DIRopen(path);

	if (dir == NULL)
	{
		return -1;
	}

	i = 0;

	while ((fname = DIRread(dir, d)) != NULL)
	{
		if (fname[0] == '.')
		{
			/* skip hidden directories ... */

			if (fname[1] == '\0')
			{
				continue; /* => "." */
			}

			if (fname[1] == '.' && fname[2] == '\0')
			{
				continue; /* => ".." */
			}

			if (pattern[0] != '.')
			{
				continue; /* ... if not wanted */
			}
		}

		if (fnmatch(pattern, fname, 0) == 0)
		{
			char		*sptr;
			const char	*separator;
			struct stat	statbuf;

			separator = dir_separator;

			sptr = g_strconcat(path, separator, fname, NULL);

#if defined(BUILD_MINGW) && defined(HAVE_GDIROPEN)
			{ char *sptr_utf8 = sptr;
			sptr = CHARCONV_UTF8_TO_FILENAME(sptr);
			if (sptr == NULL)
			{
				sptr = sptr_utf8;
			}
			else if (sptr_utf8 != sptr)
			{
				g_free(sptr_utf8);
			} }
#endif
			if (stat(sptr, &statbuf) == -1 ||
						! S_ISDIR(statbuf.st_mode))
			{
				separator = "";
			}

			g_free(sptr);

			buf[i] = g_strconcat(fname, separator, NULL);
			i++;

			if (i >= TL_FNMATCH_NAMES)
			{
				errno = ENOMEM;
				break;
			}
		}

	} /* while ((fname = g_dir_read_name(dir)) != NULL) */

	DIRclose(dir);

	if (i > 0)
	{
		data->fnamec	= i;
		buf[i]		= NULL;

		tl_sort_strarray(buf, i);
		data->fnamev = g_memdup(buf, sizeof(char*) * (i + 1));

		if (i >= TL_FNMATCH_NAMES)
		{
			return -1;
		}
	}

	return i;
}

#endif /* ! HAVE_GLOB */

CompletionData *tl_fname_completion(const gchar *fname)
{
	static const char esc_chars[] =
#ifdef BUILD_MINGW
				"*?[]";
#else
				"*?[]\\";
#endif
	int		len;
	size_t		idx, path_len, pattern_len;
	char		*beg, *str;
	CompletionData	*retv;
	char		buf[4096];

	errno = 0;

	if (fname == NULL)
	{
		return NULL;
	}

#ifdef BUILD_MINGW
	fname = g_strdup(fname);
	file_path_unix_to_dos((char *) fname);
#endif
	str = strrchr(fname, G_DIR_SEPARATOR);

	pattern_len = strlen(str == NULL ? fname : &str[1]);

	idx = 0;
	path_len = 0;

	if (fname[0] == '~' && fname[1] == G_DIR_SEPARATOR)
	{
		const char *home = g_get_home_dir();

		if (home)
		{

#if defined(BUILD_MINGW) && ! defined(HAVE_GDIROPEN)
			const char *home_utf8 = home;
			home = CHARCONV_UTF8_TO_FILENAME(home_utf8);
#endif

#ifdef HAVE_GLOB
			len = tl_str_escape_chars(buf,
					sizeof(buf) - 1, home, esc_chars);
#else
			len = strlen(home);

			if (len >= sizeof(buf) - 1)
			{
				len = -1;
			}

			else
			{
				memcpy(buf, home, len);
			}
#endif

#if defined(BUILD_MINGW) && ! defined(HAVE_GDIROPEN)
			CHARCONV_UTF8_FREE((char *) home, home_utf8);
#endif
			if (len < 0)
			{
#ifdef BUILD_MINGW
				g_free((char *) fname);
#endif
				return NULL;
			}

			idx = 2;
			buf[len] = G_DIR_SEPARATOR;
			path_len = len + 1;
		}
	}

#ifdef HAVE_GLOB
	len = tl_str_escape_chars(&buf[path_len],
			sizeof(buf) - 1 - path_len, &fname[idx], esc_chars);
#else
	/*
		Insert escape characters only in the
		filename, needed for 'fnmatch'.
		The pathname is used for open this
		directory, so without escapes.
	*/

	len = strlen(&fname[idx]) - pattern_len;

	if (len >= sizeof(buf) - 1 - path_len)
	{
		len = -1;
	}

	else
	{
		char *s = memcpy(&buf[path_len], &fname[idx], len);

		if (pattern_len > 0)
		{
			int l = tl_str_escape_chars(&s[len],
				sizeof(buf) - 1 - path_len - len,
					&fname[idx + len], esc_chars);

			if (l < 0)
			{
				len = l;
			}

			else
			{
				len += l;
			}
		}

	} /* if (len >= sizeof(buf) - 1 - path_len) */

#endif /* #ifdef HAVE_GLOB .. #else */

#ifdef BUILD_MINGW
	g_free((char *) fname);
#endif
	if (len < 0)
	{
		return NULL;
	}

	path_len += len;
	buf[path_len] = '*';
	buf[path_len + 1] = '\0';

	retv = g_malloc(sizeof(CompletionData));

	retv->fnamec = 0;
	retv->fnamev = NULL;
	retv->append = NULL;

#ifdef HAVE_GLOB

	retv->globdata = g_malloc(sizeof(glob_t));

	if (glob(buf, (GLOB_ERR | GLOB_MARK), NULL, retv->globdata) ||
				((glob_t *) retv->globdata)->gl_pathc == 0)
#else
	if (tl_fnmatch_completion(buf, retv) <= 0)
#endif
	{
		tl_completion_free(retv);
		return NULL;
	}

#ifdef HAVE_GLOB

	/* copy pointers to filenames, skip the path */

	beg = ((glob_t *) retv->globdata)->gl_pathv[0];
	str = strrchr(beg, G_DIR_SEPARATOR);

	if (str == NULL)
	{
		/* filename */
		path_len = 0;
	}

	else
	{
		if (str[1] == '\0')
		{
			/* directory */
			while (*(--str) != G_DIR_SEPARATOR)
			{
				if (str <= beg)
				{
					break;
				}
			}
		}

		if (*(str) == G_DIR_SEPARATOR)
		{
			/* skip slash */
			str++;
		}

		path_len = str - beg;
	}

	retv->fnamec = ((glob_t *) retv->globdata)->gl_pathc;
	retv->fnamev = g_malloc((retv->fnamec + 1) * sizeof(char*));

	idx = 0;

	do
	{
		str = ((glob_t *) retv->globdata)->gl_pathv[idx];
		retv->fnamev[idx] = &str[path_len];

	}
	while (++idx < retv->fnamec);

	retv->fnamev[idx] = NULL;

#endif /* HAVE_GLOB */

	/*
		if single match, 'append' points
		to start of new characters
	*/

	retv->append = retv->fnamev[0] + pattern_len;

	if (retv->fnamec > 1)
	{
		/*
			else 'append' points to a copy of the
			first matching characters from start
		*/

		int pos;

		pos = pattern_len;
		beg = retv->fnamev[0];

		for (;;)
		{
			int	chr;
			char	**array;

			chr   = beg[pos];
			array = &retv->fnamev[1];

			while ((str = *(array++)) != NULL)
			{
				if (str[pos] != chr)
				{
					goto __done;
				}
			}

			pos++;
		}
__done:
		retv->append = g_strndup(retv->append, pos - pattern_len);
	}

	return retv;

} /* CompletionData *tl_fname_completion(const gchar*) */

void tl_completion_free(CompletionData *ptr)
{
	if (ptr)
	{

#ifdef HAVE_GLOB
		globfree(ptr->globdata);
		g_free(ptr->globdata);
#else
		if (ptr->fnamev != NULL)
		{
			char **p;

			if ((p = ptr->fnamev) != NULL)
			{
				char *f;

				while ((f = *(p++)) != NULL)
				{
					g_free(f);
				}
			}
		}
#endif
		g_free(ptr->fnamev);

		if (ptr->fnamec > 1)
		{
			g_free(ptr->append);
		}

		g_free(ptr);
	}

} /* void tl_completion_free(CompletionData*) */

#endif /* CONFIG_FNAME_COMPLETION */

#ifdef CONFIG_BERKELEY_DB

char *tl_strnrsep(char *str, int len, int *sep)
{
	int c = *sep;

	if (c && len > 0)
	{
		int  n;
		char *s;

		if (c < 0)
		{
			static const char sep_str[] = "/\\:";

#ifdef PRE_INCREMENT
			const char *p = sep_str - 1;
			while (s = &str[len], (c = *(++p)) != '\0')
#else
			const char *p = sep_str;
			while (s = &str[len], (c = *(p++)) != '\0')
#endif
			{
				while (s > str)
				{
					if (*(--s) == c)
					{
						break;
					}
				}

				if (s > str)
				{
					break;
				}
			}

			*sep = c;
		}

		n = 2;
		s = &str[len];

		while (s > str)
		{
			if (*(--s) == c)
			{
				*s = '/';

				if (--n <= 0)
				{
					return &s[1];
				}
			}
		}

	} /* if (c && len > 0) */

	return str;

} /* char *tl_strnrsep(char*, int, int*) */

#endif /* CONFIG_BERKELEY_DB */

#if GLIB_MAJOR_VERSION == 1

char *tl_strchr_nth(const char *str, int chr, int n)
{
	str--;

	while ((str = strchr(++str, chr)) != NULL)
	{
		if (--n == 0)
		{
			break;
		}
	}

	return (char *) str;

} /* char *tl_strchr_nth(const char*, int, int) */

#endif /* GLIB_MAJOR_VERSION == 1 */

/* EOF */
